/*
 * UbuntuOne Nautilus plugin
 *
 * Authors: Rodrigo Moya <rodrigo.moya@canonical.com>
 *
 * Copyright 2009-2010 Canonical Ltd.
 *
 * This program is free software: you can redistribute it and/or modify it
 * under the terms of the GNU General Public License version 3, as published
 * by the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranties of
 * MERCHANTABILITY, SATISFACTORY QUALITY, or FITNESS FOR A PARTICULAR
 * PURPOSE.  See the GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License along
 * with this program.  If not, see <http://www.gnu.org/licenses/>.
 *
 */

#include <glib/gi18n.h>
#include <libsyncdaemon/syncdaemon-folders-interface.h>
#include <libsyncdaemon/syncdaemon-shares-interface.h>
#include "location-widget.h"

#define EXPANDER_STATUS_SECTION "/apps/ubuntuone/nautilus"
#define EXPANDER_STATUS_KEY EXPANDER_STATUS_SECTION "/expand-location"

G_DEFINE_TYPE(LocationWidget, location_widget, GTK_TYPE_HBOX)

static void
set_label_text (LocationWidget *location, gboolean enabled)
{
	gchar *labeltext;

	if (enabled) {
		labeltext = g_strdup_printf ("<b>Ubuntu One</b> - %s", _("folder synchronized"));
		gtk_widget_hide (location->help_label);
	} else {
		labeltext = g_strdup_printf ("<b>Ubuntu One</b> - %s", _("folder not synchronized"));
		if (location->expander_status == GTK_ARROW_DOWN)
			gtk_widget_show (location->help_label);
	}
		
	gtk_label_set_markup (GTK_LABEL (location->info_label), labeltext);
	g_free (labeltext);
}

static void
start_spinner (LocationWidget *location)
{
	gtk_widget_hide (location->help_label);
	gtk_widget_show (location->spinner);
	gtk_spinner_start (GTK_SPINNER (location->spinner));
}

static void
stop_spinner (LocationWidget *location)
{
	gtk_spinner_stop (GTK_SPINNER (location->spinner));
	gtk_widget_hide (location->spinner);
}

static void
folder_created_cb (SyncdaemonDaemon *daemon, gboolean success, SyncdaemonFolderInfo *folder_info, gpointer user_data)
{
	const gchar *path;
	LocationWidget *location = LOCATION_WIDGET (user_data);

	path = syncdaemon_folder_info_get_path (folder_info);
	g_warning ("in folder_created_cb with path = %s", path);
	if (g_strcmp0 (path, location->path) == 0) {
		set_label_text (location, success);
		if (success) {
			gtk_widget_hide (location->expander);
			gtk_widget_set_sensitive (location->toggle_button, TRUE);
			gtk_widget_hide (location->help_label);
		} else {
			gtk_widget_set_sensitive (location->toggle_button, TRUE);
			ubuntuone_show_error_dialog (location->uon, _("Error enabling folder"),
						     _("Could not enable folder %s for synchronizing to Ubuntu One"),
						     location->path);
		}

		stop_spinner (location);
	}
}

static void
folder_deleted_cb (SyncdaemonDaemon *daemon, gboolean success, SyncdaemonFolderInfo *folder_info, gpointer user_data)
{
	const gchar *path;
	LocationWidget *location = LOCATION_WIDGET (user_data);

	path = syncdaemon_folder_info_get_path (folder_info);
	if (g_strcmp0 (path, location->path) == 0) {
		if (success) {
			gtk_widget_show (location->expander);
			set_label_text (location, FALSE);
			gtk_widget_set_sensitive (location->toggle_button, TRUE);
			if (location->expander_status == GTK_ARROW_DOWN)
				gtk_widget_show (location->help_label);
		} else {
			set_label_text (location, TRUE);
			gtk_widget_set_sensitive (location->toggle_button, TRUE);
			ubuntuone_show_error_dialog (location->uon, _("Error disabling folder"),
						     _("Could not disable folder %s for synchronizing to Ubuntu One"),
						     location->path);
		}

		stop_spinner (location);
	}
}

static void
location_widget_finalize (GObject *object)
{
	LocationWidget *location = LOCATION_WIDGET (object);

	g_signal_handlers_disconnect_by_func (location->uon->syncdaemon, folder_created_cb, location);
	g_signal_handlers_disconnect_by_func (location->uon->syncdaemon, folder_deleted_cb, location);

	if (location->conf_notify_id != 0)
		gconf_client_notify_remove (location->conf_client, location->conf_notify_id);

	if (location->conf_client != NULL)
		g_object_unref (G_OBJECT (location->conf_client));

	if (location->path != NULL)
		g_free (location->path);

	G_OBJECT_CLASS (location_widget_parent_class)->finalize (object);
}

static void
location_widget_class_init (LocationWidgetClass *klass)
{
	GObjectClass *object_class = G_OBJECT_CLASS (klass);

	object_class->finalize = location_widget_finalize;
}

static gboolean
check_shared_for_folder (LocationWidget *location, SyncdaemonFolderInfo *folder_info)
{
	SyncdaemonInterface *interface;
	GHashTable *shares = NULL;
	gboolean result = TRUE;

	interface = syncdaemon_daemon_get_shares_interface (location->uon->syncdaemon);
	if (SYNCDAEMON_IS_SHARES_INTERFACE (interface)) {
		GSList *shared_list, *l;

		shared_list = syncdaemon_shares_interface_get_shared (SYNCDAEMON_SHARES_INTERFACE (interface));
		for (l = shared_list; l != NULL; l = l->next) {
			SyncdaemonShareInfo *share_info = SYNCDAEMON_SHARE_INFO (shared_list->data);

			if (g_str_has_prefix (syncdaemon_share_info_get_path (share_info),
					      syncdaemon_folder_info_get_path (folder_info))
			    && syncdaemon_share_info_get_accepted (share_info)) {
				gchar *path;
				const gchar *username;
				GSList *users = NULL;

				if (shares == NULL)
					shares = g_hash_table_new (g_str_hash, g_str_equal);

				g_hash_table_lookup_extended (shares,
							      syncdaemon_share_info_get_path (share_info),
							      (gpointer *) &path,
							      (gpointer *) &users);

				/* Add the share and the user it's shared with to the hash table */
				if (syncdaemon_share_info_get_user_visible_name (share_info) != NULL)
					username = syncdaemon_share_info_get_user_visible_name (share_info);
				else
					username = syncdaemon_share_info_get_username (share_info);

				users = g_slist_append (users, (gpointer) username);
				g_hash_table_insert (shares,
						     (gpointer) syncdaemon_share_info_get_path (share_info),
						     (gpointer) users);
			}
		}

		g_slist_free (shared_list);

		/* If there are shares, ask the user what to do */
		if (shares != NULL) {
			GtkWidget *dialog;
			GHashTableIter iter;
			GString *question;
			gchar *path;
			GSList *users;

			question = g_string_new (_("This folder contains folders that have been shared:\n"));

			g_hash_table_iter_init (&iter, shares);
			while (g_hash_table_iter_next (&iter, (gpointer) &path, (gpointer) &users)) {
				gint count = 0;

				question = g_string_append (question, "\t- ");
				question = g_string_append (question, path);
				question = g_string_append (question, " (");

				while (users != NULL) {
					gchar *username = (gchar *) users->data;

					if (count > 0)
						question = g_string_append (question, ", ");
					question = g_string_append (question, username);

					users = g_slist_remove (users, username);
					count += 1;
				}

				question = g_string_append (question, ")\n");
			}

			question = g_string_append (question, _("This action will un-share these items. Other users"
								" will no longer be able to access these files."
								" Would you like to proceed?"));

			dialog = gtk_message_dialog_new (
				GTK_WINDOW (gtk_widget_get_toplevel (GTK_WIDGET (location))),
				0, GTK_MESSAGE_QUESTION,
				GTK_BUTTONS_YES_NO,
				"%s", question->str);
			if (gtk_dialog_run (GTK_DIALOG (dialog)) == GTK_RESPONSE_YES)
				result = FALSE;

			gtk_widget_destroy (dialog);
			g_string_free (question, TRUE);
			g_hash_table_destroy (shares);
		}

	}

	return result;
}

static void
disable_u1 (LocationWidget *location)
{
	SyncdaemonInterface *interface;

	/* Perform the removal of this folder */
	interface = syncdaemon_daemon_get_folders_interface (location->uon->syncdaemon);
	if (interface != NULL) {
		SyncdaemonFolderInfo *folder_info;

		folder_info = syncdaemon_folders_interface_get_info (
			SYNCDAEMON_FOLDERS_INTERFACE (interface),
			location->path);
		if (folder_info != NULL && check_shared_for_folder (location, folder_info)) {
			set_label_text (location, FALSE);
			gtk_widget_set_sensitive (location->toggle_button, FALSE);

			syncdaemon_folders_interface_delete (
				SYNCDAEMON_FOLDERS_INTERFACE (interface),
				syncdaemon_folder_info_get_volume_id (folder_info));

			g_object_unref (G_OBJECT (folder_info));
		}
	}
}

static void
enable_u1 (LocationWidget *location)
{
	SyncdaemonInterface *interface;

	/* Perform the addition of this folder */
	interface = syncdaemon_daemon_get_folders_interface (location->uon->syncdaemon);
	if (interface != NULL) {
		gchar *labeltext;

		/* Update label to show progress */
		labeltext = g_strdup_printf ("<b>Ubuntu One</b> - %s", _("sync in progress"));
		gtk_label_set_markup (GTK_LABEL (location->info_label), labeltext);
		g_free (labeltext);

		start_spinner (location);		
		gtk_widget_set_sensitive (location->toggle_button, FALSE);

		syncdaemon_folders_interface_create (SYNCDAEMON_FOLDERS_INTERFACE (interface),
						     location->path);
	}
}

static void
sync_toggled_cb (GtkToggleButton *button, gpointer user_data)
{
	LocationWidget *location = LOCATION_WIDGET (user_data);

	if (gtk_toggle_button_get_active (GTK_TOGGLE_BUTTON (location->toggle_button)))
		enable_u1 (location);
	else
		disable_u1 (location);
}

static gboolean
expander_clicked_cb (GtkWidget *widget, GdkEventButton *event, gpointer user_data)
{
	LocationWidget *location = LOCATION_WIDGET (user_data);

	if (event->button == 1) {
		switch (location->expander_status) {
		case GTK_ARROW_DOWN: /* expanded */
			gconf_client_set_bool (location->conf_client, EXPANDER_STATUS_KEY, FALSE, NULL);
			break;
		case GTK_ARROW_RIGHT: /* collapsed */
			gconf_client_set_bool (location->conf_client, EXPANDER_STATUS_KEY, TRUE, NULL);
			break;
		}
	}

	return FALSE;
}

static void
config_changed_cb (GConfClient *client, guint cnx_id, GConfEntry *entry, gpointer user_data)
{
	LocationWidget *location = LOCATION_WIDGET (user_data);

	if (g_strcmp0 (gconf_entry_get_key (entry), EXPANDER_STATUS_KEY) == 0) {
		if (gconf_value_get_bool (gconf_entry_get_value (entry))) {
			location->expander_status = GTK_ARROW_DOWN;
			gtk_widget_show (location->help_label);
		} else {
			location->expander_status = GTK_ARROW_RIGHT;
			gtk_widget_hide (location->help_label);
		}

		gtk_arrow_set (GTK_ARROW (location->expander),  location->expander_status, GTK_SHADOW_IN);
	}
}

static void
location_widget_init (LocationWidget *location)
{
	GtkWidget *table, *box;

	location->path = NULL;
	location->info_label = NULL;

	/* Read expander status from GConf */
	location->conf_client = gconf_client_get_default ();
	gconf_client_add_dir (location->conf_client, EXPANDER_STATUS_SECTION, GCONF_CLIENT_PRELOAD_RECURSIVE, NULL);
	location->conf_notify_id = gconf_client_notify_add (location->conf_client,
							    EXPANDER_STATUS_SECTION,
							    (GConfClientNotifyFunc) config_changed_cb,
							    location,
							    NULL,
							    NULL);

	if (gconf_client_get_bool (location->conf_client, EXPANDER_STATUS_KEY, NULL))
		location->expander_status = GTK_ARROW_DOWN;
	else
		location->expander_status = GTK_ARROW_RIGHT;

	/* Create the widgets in the interface */
	table = gtk_table_new (2, 4, FALSE);
	gtk_widget_show (table);
	gtk_box_pack_start (GTK_BOX (location), table, TRUE, TRUE, 0);

	box = gtk_event_box_new ();
	g_signal_connect (G_OBJECT (box), "button-press-event",
			  G_CALLBACK (expander_clicked_cb), location);
	gtk_widget_show (box);
	gtk_table_attach (GTK_TABLE (table), box, 0, 1, 0, 1,
			  GTK_FILL, GTK_FILL, 3, 3);

	location->expander = gtk_arrow_new (location->expander_status, GTK_SHADOW_IN);
	gtk_widget_show (location->expander);
	gtk_container_add (GTK_CONTAINER (box), location->expander);

	location->info_label = gtk_label_new ("<b>Ubuntu One</b>");
	gtk_label_set_use_markup (GTK_LABEL (location->info_label), TRUE);
	gtk_misc_set_alignment (GTK_MISC (location->info_label), 0.0, 0.5);
	gtk_widget_show (location->info_label);
	gtk_table_attach (GTK_TABLE (table), location->info_label, 1, 2, 0, 1,
			  GTK_EXPAND | GTK_FILL | GTK_SHRINK, GTK_FILL, 3, 3);

	location->help_label = gtk_label_new (_("These files are not backed up and will not be available in your other computers"));
	gtk_label_set_use_markup (GTK_LABEL (location->help_label), TRUE);
	gtk_misc_set_alignment (GTK_MISC (location->help_label), 0.0, 0.5);
	gtk_table_attach (GTK_TABLE (table), location->help_label, 1, 3, 1, 2,
			  GTK_EXPAND | GTK_FILL | GTK_SHRINK, GTK_FILL, 3, 3);

	location->spinner = gtk_spinner_new ();
	gtk_table_attach (GTK_TABLE (table), location->spinner, 2, 3, 0, 1,
			  GTK_SHRINK, GTK_SHRINK, 3, 3);

	/* Buttons */
	location->toggle_button = gtk_check_button_new_with_label (_("Synchronize folder"));
	gtk_widget_show (location->toggle_button);
	gtk_table_attach (GTK_TABLE (table), location->toggle_button, 3, 4, 0, 1,
			  GTK_FILL, GTK_FILL, 3, 3);
}

static gboolean
is_special_udf (UbuntuOneNautilus *uon, const gchar *path)
{
	gint i;
	gchar *special_udfs[] = {
		"Ubuntu One",
		"Ubuntu One/Shared With Me",
		".ubuntuone/Purchased from Ubuntu One",
		".local/share/ubuntuone"
	};

	for (i = 0; i < G_N_ELEMENTS (special_udfs); i++) {
		gchar *tmp;
		gboolean done = FALSE;

		tmp = g_build_filename (g_get_home_dir (), special_udfs[i], NULL);
		if (g_str_has_prefix (path, tmp))
			done = TRUE;

		g_free (tmp);

		if (done)
			return TRUE;
	}

	return FALSE;
}

GtkWidget *
location_widget_new (UbuntuOneNautilus *uon, const gchar *path)
{
	LocationWidget *location;
	gboolean is_root;

	g_return_val_if_fail (uon != NULL, NULL);
	g_return_val_if_fail (path != NULL, NULL);

	/* We only show the location bar for user's folders */
	if (!g_str_has_prefix (path, g_get_home_dir ()) ||
	    strlen (path) == strlen (g_get_home_dir ()))
		return NULL;

	location = (LocationWidget *) g_object_new (TYPE_LOCATION_WIDGET, NULL);
	location->uon = uon;
	location->path = g_strdup (path);

	/* Connect to Syncdaemon signals */
	g_signal_connect (G_OBJECT (uon->syncdaemon), "folder_created",
			  G_CALLBACK (folder_created_cb), location);
	g_signal_connect (G_OBJECT (uon->syncdaemon), "folder_deleted",
			  G_CALLBACK (folder_deleted_cb), location);

	if (ubuntuone_is_storagefs (uon, path, &is_root)) {
		/* Create label and disable button */
		gtk_widget_hide (location->expander);
		set_label_text (location, TRUE);
		gtk_toggle_button_set_active (GTK_TOGGLE_BUTTON (location->toggle_button), TRUE);
		if (!is_root || is_special_udf (uon, path))
			gtk_widget_set_sensitive (location->toggle_button, FALSE);
	} else {
		/* Check if this is the 'Shared with me' folder */
		if (is_special_udf (uon, path)) {
			gtk_widget_hide (location->expander);
			set_label_text (location, TRUE);
			gtk_toggle_button_set_active (GTK_TOGGLE_BUTTON (location->toggle_button), TRUE);
			gtk_widget_set_sensitive (location->toggle_button, FALSE);
		} else {
			GSList *udfs, *l;
			SyncdaemonInterface *interface;
			gboolean allow_enabling = TRUE;

			/* Check to see if this is the parent of a UDF */
			interface = syncdaemon_daemon_get_folders_interface (uon->syncdaemon);
			udfs = syncdaemon_folders_interface_get_folders (SYNCDAEMON_FOLDERS_INTERFACE (interface));
			for (l = udfs; l != NULL; l = l->next) {
				SyncdaemonFolderInfo *folder_info = SYNCDAEMON_FOLDER_INFO (l->data);

				if (g_str_has_prefix (syncdaemon_folder_info_get_path (folder_info), path))
					allow_enabling = FALSE;
			}

			g_slist_free (udfs);
			
			gtk_toggle_button_set_active (GTK_TOGGLE_BUTTON (location->toggle_button), FALSE);
			if (allow_enabling) {
				set_label_text (location, FALSE);
			} else {
				gchar *labeltext;

				labeltext = g_strdup_printf ("<b>Ubuntu One</b> - %s",
							     _("cannot synchronize this folder"));
				gtk_label_set_markup (GTK_LABEL (location->info_label), labeltext);
				g_free (labeltext);

				gtk_label_set_text (GTK_LABEL (location->help_label),
						    _("This folder cannot be synchronized because it contains "
						      "one or more folders that already synchronized"));
				if (location->expander_status == GTK_ARROW_DOWN)
					gtk_widget_show (location->help_label);

				gtk_toggle_button_set_active (GTK_TOGGLE_BUTTON (location->toggle_button), FALSE);
				gtk_widget_set_sensitive (location->toggle_button, FALSE);
			}
		}
	}

	g_signal_connect (G_OBJECT (location->toggle_button), "toggled",
			  G_CALLBACK (sync_toggled_cb), location);

	gtk_widget_show (GTK_WIDGET (location));

	return GTK_WIDGET (location);
}
